
NUMROWS           equ   3
NUMCOLUMNS        equ   7
NUMINVADERS       equ   NUMROWS*NUMCOLUMNS

; offsets into invader structure
INV_X             equ   0     ; chars across
INV_Y             equ   1     ; pixels down
INV_DEAD          equ   2     ; non-zero if invader is dead
INV_GFXLOW        equ   3     ; ptr to alien gfx data
INV_GFXHIGH       equ   4
INV_COLOUR        equ   5     ; colour
INV_SHOT          equ   6     ; has this invader shot?
INV_SHOTX         equ   7     ; chars across of bullet
INV_SHOTY         equ   8     ; pixels down of bullet
INV_SHOTSPD       equ   9     ; speed of bullet
INV_NUMBER        equ   10    ; invader number
INV_SCORE         equ   11    ; score for this invader

BYTESPERINVADER   equ   12    ; bytes for each invader structure


invgfxoffsettype  db    0
invgfxoffsetptr   dw    0
invcolour         db    0
invadernum        db    0
invaderscore      db    0

SetupInvaders:    xor   a
                  ld    (invgfxoffsettype),a
                  ld    (invadernum),a
                  ld    a,7
                  ld    (invcolour),a
                  ld    (invaderscore),a

                  ld    ix,InvaderTable
                  ld    hl,#0000

                  ld    a,(Level)
                  dec   a                 ; make zero based
                  and   7
                  cp    7
                  jr    nz,invnotmax
                  dec   a

invnotmax:        add   a,a               ; multiply by 8 to bring them down 1 char per level
                  add   a,a
                  add   a,a
                  add   a,h
                  ld    h,a

                  ld    b,NUMROWS
StInv1:           push  bc
                  push  hl

                  push  hl
                  ld    a,(invgfxoffsettype)
                  ld    h,0
                  ld    l,a
                  add   hl,hl ; multiply by 48
                  add   hl,hl
                  add   hl,hl
                  add   hl,hl
                  ld    d,h
                  ld    e,l
                  add   hl,hl
                  add   hl,de

                  add   hl,hl
                  add   hl,hl ; multiply by 4 (number of frames of 48 bytes)
                  ld    de,invadergfx
                  add   hl,de
                  ld    (invgfxoffsetptr),hl
                  pop   hl

                  ld    a,(invaderscore)
                  sub   2
                  ld    (invaderscore),a

                  ld    b,NUMCOLUMNS
StInv2:           ld    (ix+INV_X),l
                  ld    (ix+INV_Y),h

                  xor   a
                  ld    (ix+INV_DEAD),a
                  ld    (ix+INV_SHOT),a

                  ld    de,(invgfxoffsetptr)
                  ld    (ix+INV_GFXLOW),e
                  ld    (ix+INV_GFXHIGH),d

                  ld    a,(invcolour)
                  ld    (ix+INV_COLOUR),a

                  ld    a,(invadernum)
                  ld    (ix+INV_NUMBER),a

                  ld    a,(invaderscore)
                  ld    (ix+INV_SCORE),a

                  ld    de,BYTESPERINVADER
                  add   ix,de
                  inc   l
                  inc   l
                  djnz  StInv2

                  ld    hl,invgfxoffsettype
                  inc   (hl)
                  ld    hl,invadernum
                  inc   (hl)

                  ld    a,(invcolour)
                  sub   2
                  ld    (invcolour),a
                  
                  pop   hl
                  ld    a,h
                  add   a,24  ; down 24 pixels
                  ld    h,a
                  pop   bc
                  djnz  StInv1

                  ld    a,NUMINVADERS
                  ld    (invadersalive),a
                  ld    a,1
                  ld    (invdir),a  ; invaders go right
                  ld    a,3
                  ld    (invframe),a      ; starting frame
                  xor   a
                  ld    (flipinvdir),a
                  ld    (invlanded),a
                  ret

GetInvaderGfx:    ld    a,(invframe)
                  ld    h,0
                  ld    l,a
                  add   hl,hl ; multiply by 48
                  add   hl,hl
                  add   hl,hl
                  add   hl,hl
                  ld    d,h
                  ld    e,l
                  add   hl,hl
                  add   hl,de
                  ex    de,hl
                  ret

MoveInvaders:     call  InvaderTone

                  ld    a,(invadersalive)
                  or    a
                  rra
                  ld    (InvaderTimer),a

                  ld    a,(flipinvdir)
                  or    a
                  jp    z,handleinvnormal
                  
                  xor   a
                  ld    (flipinvdir),a
                  ld    a,1
                  ld    (moveinvdown),a
                  jp    goinvaders

handleinvnormal:  xor   a
                  ld    (moveinvdown),a
                  ld    (invchardir),a

                  ld    a,(invdir)
                  or    a
                  jp    nz,invgoright

invgoleft:        ld    a,(invframe)
                  inc   a
                  and   3
                  ld    (invframe),a
                  jp    nz,goinvaders
                  
                  ld    a,-1
                  ld    (invchardir),a
                  jp    goinvaders

invgoright:       ld    a,(invframe)
                  dec   a
                  and   3
                  ld    (invframe),a
                  cp    3
                  jp    nz,goinvaders
                  
                  ld    a,1
                  ld    (invchardir),a

goinvaders:       ld    ix,InvaderTable
                  ld    b,NUMINVADERS

HnInv1:           ld    a,(ix+INV_DEAD)
                  or    a
                  jp    nz,Hn_Nxt
                  
                  push  bc
                  call  MoveInvader
                  pop   bc

Hn_Nxt:           ld    de,BYTESPERINVADER
                  add   ix,de
                  dec   b
                  jp    nz,HnInv1
                  
                  ld    a,(moveinvdown)
                  or    a
                  ret   z
                  
                  ld    a,(invdir)
                  xor   1
                  ld    (invdir),a
                  xor   a
                  ld    (moveinvdown),a
                  ret

MoveInvader:      ld    a,(moveinvdown)
                  or    a
                  jp    nz,movedownrow
                  
                  ld    a,(invchardir)
                  or    a
                  ret   z
                  
                  ld    c,a
                  ld    a,(ix+INV_X)
                  add   a,c
                  ld    (ix+INV_X),a
                  and   31
                  jp    z,invlimit
                  cp    29
                  jp    z,invlimit

                  ; see if this invader can shoot
                  ld    a,(ix+INV_SHOT)
                  or    a
                  ret   nz

                  ld    a,(invshoottimer)
                  inc   a
                  ld    (invshoottimer),a
                  and   15
                  ret   nz

                  ld    (invader_IX),ix

                  ld    a,(ix+INV_NUMBER)

checkinvaderbelow:
                  add   a,NUMCOLUMNS
                  cp    NUMINVADERS
                  jr    nc,SetInvaderShoot

                  ld    de,NUMCOLUMNS*BYTESPERINVADER
                  add   ix,de
                  ld    c,a
                  ld    a,(ix+INV_DEAD)
                  or    a
                  ld    a,c
                  jr    nz,checkinvaderbelow

                  ld    ix,(invader_IX)
                  ret

SetInvaderShoot:  ld    ix,(invader_IX)

                  ld    a,(on_menu)
                  or    a
                  jr    z,setshoot1

                  ld    a,(ix+INV_Y)    ; on menu shooting
                  add   a,24
                  jr    setshoot2

setshoot1:        ld    a,(ix+INV_Y)    ; normal in-game shooting
                  add   a,24
                  cp    110
                  ret   nc

setshoot2:        ld    (ix+INV_SHOTY),a

                  ld    (ix+INV_SHOT),1
                  ld    a,(ix+INV_X)
                  ld    (ix+INV_SHOTX),a
                  ld    a,r
                  and   3
                  jr    nz,setshotspd
                  ld    a,2

setshotspd:       add   a,a
                  ld    (ix+INV_SHOTSPD),a
                  ret

invlimit:         ld    a,1
                  ld    (flipinvdir),a
                  ret

movedownrow:      ld    a,(on_menu)
                  or    a
                  ret   nz

                  ld    a,(ix+INV_Y)
                  add   a,8
                  ld    (ix+INV_Y),a
                  cp    128
                  ret   c     ; okay if less than 128 pixels down

                  ld    a,1
                  ld    (invlanded),a
                  ret

;--------------------------------------------------------------------------------

DrawInvaders:     call  GetInvaderGfx
                  ld    (invgfxoffset),de       ; new invader image

                  ld    ix,InvaderTable
                  ld    b,NUMINVADERS

DrawInv1:         ld    a,(ix+INV_DEAD)
                  or    a
                  jp    nz,DrawInv2

                  push  bc
                  ld    l,(ix+INV_GFXLOW)
                  ld    h,(ix+INV_GFXHIGH)
                  ld    de,(invgfxoffset)
                  add   hl,de
                  ex    de,hl
                  call  PrtInvader
                  pop   bc

DrawInv2:         ld    de,BYTESPERINVADER
                  add   ix,de
                  dec   b
                  jp    nz,DrawInv1
                  ret

;--------------------------------------------------------------------------------

; IX = address of structure for invader
; DE = address of graphic data

PrtInvader:       ld    l,(ix+INV_X)
                  ld    h,(ix+INV_Y)
                  call  MapScreenAddress
                  push  hl

                  ld    b,16
prtiv1:           ld    a,(de)
                  or    (hl)
                  ld    (hl),a
                  inc   de
                  inc   l
                  ld    a,(de)
                  or    (hl)
                  ld    (hl),a
                  inc   de
                  inc   l
                  ld    a,(de)
                  or    (hl)
                  ld    (hl),a
                  inc   de
                  dec   l
                  dec   l
                  call  NextPixelLine
                  djnz  prtiv1

                  pop   hl
                  ld    c,(ix+INV_COLOUR)
                  ld    a,h
                  rrca
                  rrca
                  rrca
                  and   3
                  or    &d8
                  ld    h,a
                  ld    (hl),c
                  inc   l
                  ld    (hl),c
                  inc   l
                  ld    (hl),c
                  ld    de,30
                  add   hl,de
                  ld    (hl),c
                  inc   l
                  ld    (hl),c
                  inc   l
                  ld    (hl),c
                  ret

HandleInvaderBullets:
                  ld    ix,InvaderTable
                  ld    b,NUMINVADERS

InvBullet1:       ld    a,(ix+INV_SHOT)
                  or    a
                  jr    z,InvBulLoop

                  ld    a,(ix+INV_SHOTY)
                  add   a,(ix+INV_SHOTSPD)
                  cp    150-8
                  jr    nc,EndInvBullet
                  ld    (ix+INV_SHOTY),a

                  ld    h,a
                  ld    l,(ix+INV_SHOTX)
                  call  MapScreenAddress
                  ld    de,InvBulletGfx
                  ld    c,8
DrawInvBlt:       ld    a,(de)
                  or    (hl)
                  ld    (hl),a
                  inc   de
                  call  NextPixelLine
                  dec   c
                  jp    nz,DrawInvBlt

InvBulLoop:       ld    de,BYTESPERINVADER
                  add   ix,de
                  djnz  InvBullet1
                  ret

EndInvBullet:     ld    (ix+INV_SHOT),0
                  jp    InvBulLoop

ClearInvaderBullets:
                  ld    ix,InvaderTable
                  ld    de,BYTESPERINVADER
                  ld    b,NUMINVADERS
ClrBullLp:        ld    (ix+INV_SHOT),0
                  add   ix,de
                  djnz  ClrBullLp
                  ret

InvaderTable:     ds    NUMINVADERS*BYTESPERINVADER

InvBulletGfx:     db    %00010000
                  db    %00010000
                  db    %00010000
                  db    %00010000
                  db    %00010000
                  db    %00010000
                  db    %00111000
                  db    %00010000

invader_IX        dw    0
invgfxoffset      dw    0     ; offset to current invader graphic frame
invdir            db    0
invframe          db    0
invchardir        db    0     ; 0, 1, -1; evaluated and added to invader's SCRLOW posn per frame
flipinvdir        db    0
moveinvdown       db    0     ; non-zero when moving down a row
invadersalive     db    0     ; number of invaders still alive
invlanded         db    0
invshoottimer     db    0

